@@ -1,4 +1,4 @@
-#define MOD_SCGI_VERSION "1.2"
+#define MOD_SCGI_VERSION "1.7"
 #define SCGI_PROTOCOL_VERSION "1"
 /* #define VERBOSE_DEBUG */
 
@@ -18,15 +18,25 @@
 #define ENABLED 1
 #define DISABLED 2
 
+typedef struct {
+	char *path;
+	unsigned long addr;	/* address in network byte order */
+	unsigned short port;	/* port in network byte order */
+} mount_entry;
+
 /*
  * Configuration record.  Used per-directory configuration data.
  */
 typedef struct {
-	unsigned long addr;	/* address in network byte order */
-	unsigned short port; 	/* port in network byte order */
+	mount_entry mount;
 	int enabled;		/* mod_scgi enabled from this directory */
 } scgi_cfg;
 
+/* Server level configuration */
+typedef struct {
+	array_header *mounts;
+} scgi_server_cfg;
+
 /*
  * Declare ourselves so the configuration routines can find and know us.
  * We'll fill it in at the end of the module.
@@ -38,48 +48,94 @@ module MODULE_VAR_EXPORT scgi_module;
  */
 static scgi_cfg *our_dconfig(request_rec *r)
 {
-    return (scgi_cfg *) ap_get_module_config(r->per_dir_config, &scgi_module);
+	return (scgi_cfg *) ap_get_module_config(r->per_dir_config,
+						 &scgi_module);
+}
+
+static scgi_server_cfg *our_sconfig(server_rec *s)
+{
+    return (scgi_server_cfg *) ap_get_module_config(s->module_config,
+						    &scgi_module);
+}
+
+static int
+mount_entry_matches(char *url, char *prefix, char **path_info)
+{
+	int i;
+	for (i=0; prefix[i] != '\0'; i++) {
+		if (url[i] == '\0' || url[i] != prefix[i])
+			return 0;
+	}
+	if (url[i] == '\0' || url[i] == '/') {
+		*path_info = url + i;
+		return 1;
+	}
+	return 0;
 }
 
 static int scgi_trans(request_rec *r)
 {
 	scgi_cfg *cfg = our_dconfig(r);
-
-	if (cfg->enabled == ENABLED) {
+	scgi_server_cfg *scfg = our_sconfig(r->server);
+	if (cfg->enabled == DISABLED) {
+		return DECLINED;
+	}
+	if (cfg->mount.addr != UNSET) {
 		r->handler = "scgi-handler";
+		ap_set_module_config(r->request_config, &scgi_module,
+				     &cfg->mount);
 		return OK;
 	}
+	else {
+    		int i;
+    		mount_entry *entries = (mount_entry *) scfg->mounts->elts;
+		for (i = 0; i < scfg->mounts->nelts; ++i) {
+			char *path_info;
+			mount_entry *p = &entries[i];
+			if (mount_entry_matches(r->uri, p->path, &path_info)) {
+				r->handler = "scgi-handler";
+				r->path_info = path_info;
+				ap_set_module_config(r->request_config,
+						     &scgi_module,
+						     p);
+				return OK;
+			}
+		}
+	}
 	return DECLINED;
 }
 
-
-int open_socket(request_rec *r, scgi_cfg *cfg)
+int open_socket(request_rec *r)
 {
 	int retries, sleeptime, rv;
 	struct sockaddr_in addr;
 	int sock;
+	mount_entry *p = (mount_entry *) ap_get_module_config(r->request_config,
+							      &scgi_module);
 
-	if (cfg->addr == UNSET)
+	if (p->addr == UNSET)
 		addr.sin_addr.s_addr = inet_addr("127.0.0.1");
 	else
-		addr.sin_addr.s_addr = cfg->addr;
-	if (cfg->port == UNSET)
+		addr.sin_addr.s_addr = p->addr;
+	if (p->port == UNSET)
 		addr.sin_port = htons(4000);
 	else
-		addr.sin_port = cfg->port;
+		addr.sin_port = p->port;
 	addr.sin_family = AF_INET;
-  
+
+	/* try to connect */
+	retries = 4;
+	sleeptime = 1;
+
+restart:
 	/* create the socket */
 	sock = ap_psocket(r->pool, AF_INET, SOCK_STREAM, 0);
 	if (sock == -1) 
 		return -1;
   
-	/* try to connect */
-	retries = 4;
-	sleeptime = 1;
-restart:
 	rv = connect(sock, (struct sockaddr *)&addr, sizeof addr);
 	if (rv != 0) {
+		ap_pclosesocket(r->pool, sock);
 		if (errno == EINTR)
 			goto restart; /* signals suck */
 		if (errno == ECONNREFUSED && retries > 0) {
@@ -107,6 +163,24 @@ restart:
 	return sock;
 }
 
+static char *http2env(pool *p, const char *name)
+{
+	char *env_name = ap_pstrcat(p, "HTTP_", name, NULL);
+	char *cp;
+
+	for (cp = env_name + 5; *cp != 0; cp++) {
+		if (*cp == '-') {
+			*cp = '_';
+		}
+		else {
+			*cp = ap_toupper(*cp);
+		}
+	}
+
+	return env_name;
+}
+
+
 static char *lookup_name(table *t, const char *name)
 {
 	array_header *hdrs_arr;
@@ -183,15 +257,26 @@ static void add_header(table *t, const char *name, const char *value)
 	}
 }
 
-static int send_headers(request_rec *r, BUFF *f, scgi_cfg *cfg)
+static int find_path_info(const char *uri, const char *path_info)
 {
-	table *t = ap_make_table(r->pool, 40); /* headers to send */
-	array_header *hdrs_arr;
-	table_entry *hdrs;
+	int n;
+	n = strlen(uri) - strlen(path_info);
+	ap_assert(n >= 0);
+	return n;
+}
+
+static int send_headers(request_rec *r, BUFF *f)
+{
+	table *t;
+	array_header *hdrs_arr, *env_arr;
+	table_entry *hdrs, *env;
 	int i;
 	unsigned long n;
 
 	log_debug(r, "sending headers");
+	t = ap_make_table(r->pool, 40); /* headers to send */
+	if (!t)
+		return 0;
 	/* CONTENT_LENGTH must come first and always be present */
 	add_header(t, "CONTENT_LENGTH",
 		   	ap_psprintf(r->pool, "%ld", r->remaining));
@@ -211,34 +296,35 @@ static int send_headers(request_rec *r, BUFF *f, scgi_cfg *cfg)
 	add_header(t, "REQUEST_METHOD", r->method);
 	add_header(t, "REQUEST_URI", original_uri(r));
 	add_header(t, "QUERY_STRING", r->args ? r->args : "");
-	add_header(t, "SCRIPT_NAME", r->uri);
-	/* skip PATH_INFO, it's always empty */
+	if (r->path_info) {
+		int path_info_start = find_path_info(r->uri, r->path_info);
+		add_header(t, "SCRIPT_NAME", ap_pstrndup(r->pool, r->uri,
+							 path_info_start));
+		add_header(t, "PATH_INFO", r->path_info);
+	}
+	else {
+		/* skip PATH_INFO, don't know it */
+		add_header(t, "SCRIPT_NAME", r->uri);
+	}
 	add_header(t, "HTTPS", lookup_env(r, "HTTPS"));
 	add_header(t, "CONTENT_TYPE", lookup_header(r, "Content-type"));
 	add_header(t, "DOCUMENT_ROOT", ap_document_root(r));
-	add_header(t, "HTTP_ACCEPT", lookup_header(r, "Accept"));
-	add_header(t, "HTTP_ACCEPT_CHARSET",
-		   lookup_header(r, "Accept-Charset"));
-	add_header(t, "HTTP_ACCEPT_ENCODING",
-		   lookup_header(r, "Accept-Encoding"));
-	add_header(t, "HTTP_ACCEPT_LANGUAGE",
-		   lookup_header(r, "Accept-Language"));
-	add_header(t, "HTTP_AUTHORIZATION", lookup_header(r, "Authorization"));
-	add_header(t, "HTTP_COOKIE", lookup_header(r, "Cookie"));
-	add_header(t, "HTTP_EXPECT", lookup_header(r, "Expect"));
-	add_header(t, "HTTP_FROM", lookup_header(r, "From"));
-	add_header(t, "HTTP_HOST", lookup_header(r, "Host"));
-	add_header(t, "HTTP_IF_MATCH", lookup_header(r, "If-Match"));
-	add_header(t, "HTTP_IF_MODIFIED_SINCE",
-		   lookup_header(r, "If-Modified-Since"));
-	add_header(t, "HTTP_IF_NONE_MATCH", lookup_header(r, "If-None-Match"));
-	add_header(t, "HTTP_IF_RANGE", lookup_header(r, "If-Range"));
-	add_header(t, "HTTP_IF_UNMODIFIED_SINCE",
-		   lookup_header(r, "If-Unmodified-Since"));
-	add_header(t, "HTTP_RANGE", lookup_header(r, "Range"));
-	add_header(t, "HTTP_REFERER", lookup_header(r, "Referer"));
-	add_header(t, "HTTP_TE", lookup_header(r, "TE"));
-	add_header(t, "HTTP_USER_AGENT", lookup_header(r, "User-Agent"));
+
+        /* HTTP headers */
+	hdrs_arr = ap_table_elts(r->headers_in);
+	hdrs = (table_entry *) hdrs_arr->elts;
+	for (i = 0; i < hdrs_arr->nelts; ++i) {
+		if (!hdrs[i].key)
+			continue;
+		add_header(t, http2env(r->pool, hdrs[i].key), hdrs[i].val);
+	}
+
+        /* environment variables */
+        env_arr = ap_table_elts(r->subprocess_env);
+        env = (table_entry *)env_arr->elts;
+        for (i = 0; i < env_arr->nelts; ++i) {
+		add_header(t, env[i].key, env[i].val);
+        }
 
 	hdrs_arr = ap_table_elts(t);
 	hdrs = (table_entry *)hdrs_arr->elts;
@@ -285,7 +371,9 @@ static int scgi_handler(request_rec *r)
 	BUFF *f;
 	char *request_body = NULL;
 	const char *location;
-	scgi_cfg *cfg = our_dconfig(r);
+
+	if (strcmp(r->handler, "scgi-handler"))
+		return DECLINED;
 
 	if ((ret = ap_setup_client_block(r, REQUEST_CHUNKED_ERROR))) {
 		return ret;
@@ -294,7 +382,7 @@ static int scgi_handler(request_rec *r)
 	/* connect to scgi server */
 	ap_hard_timeout("scgi connect", r);
 	log_debug(r, "connecting to server");
-	sock = open_socket(r, cfg);
+	sock = open_socket(r);
 	if (sock == -1) {
 		if (request_body)
 			free(request_body);
@@ -308,7 +396,7 @@ static int scgi_handler(request_rec *r)
 	ap_hard_timeout("scgi sending request", r);
 
 	/* send headers */
-        if (!send_headers(r, f, cfg)) {
+        if (!send_headers(r, f)) {
 		log_err(r, "error sending response headers");
 		return SERVER_ERROR;
 	}
@@ -322,9 +410,20 @@ static int scgi_handler(request_rec *r)
 	ap_kill_timeout(r);
 
 	log_debug(r, "reading response headers");
-	if (ap_scan_script_header_err_buff(r, f, NULL) != 0) {
-		log_err(r, "error reading response headers");
-		return SERVER_ERROR;
+	ret = ap_scan_script_header_err_buff(r, f, NULL);
+	if (ret) {
+		if (ret == SERVER_ERROR) {
+			log_err(r, "error reading response headers");
+		}
+		else {
+			/* Work around an Apache bug whereby the returned
+			 * status is ignored and status_line is used instead.
+			 * This bug is present at least in Apache 1.3.33.
+			 */
+			r->status_line = NULL;
+		}
+		ap_bclose(f);
+		return ret;
 	}
 
 	location = ap_table_get(r->headers_out, "Location");
@@ -363,13 +462,14 @@ static void scgi_init(server_rec *s, pool *p)
     ap_add_version_component("mod_scgi/" MOD_SCGI_VERSION);
 }
 
-static void *scgi_create_dir_config(pool *p, char *dirspec)
+static void *
+create_dir_config(pool *p, char *dirspec)
 {
 	scgi_cfg *cfg = (scgi_cfg *) ap_pcalloc(p, sizeof(scgi_cfg));
 
 	cfg->enabled = UNSET;
-	cfg->addr = UNSET;
-	cfg->port = UNSET;
+	cfg->mount.addr = UNSET;
+	cfg->mount.port = UNSET;
 
 	return (void *) cfg;
 }
@@ -377,7 +477,7 @@ static void *scgi_create_dir_config(pool *p, char *dirspec)
 #define MERGE(b, n, a) (n->a == UNSET ? b->a : n->a)
 
 static void *
-scgi_merge_dir_config(pool *p, void *basev, void *newv)
+merge_dir_config(pool *p, void *basev, void *newv)
 {
 	scgi_cfg *base, *new;
 	scgi_cfg *cfg = (scgi_cfg *) ap_pcalloc(p, sizeof(scgi_cfg));
@@ -385,12 +485,35 @@ scgi_merge_dir_config(pool *p, void *basev, void *newv)
 	new = (scgi_cfg *) newv;
 	
 	cfg->enabled = MERGE(base, new, enabled);
-	cfg->addr = MERGE(base, new, addr);
-	cfg->port = MERGE(base, new, port);
+	cfg->mount.addr = MERGE(base, new, mount.addr);
+	cfg->mount.port = MERGE(base, new, mount.port);
 
 	return (void *) cfg;
 }
 
+static void *
+create_server_config(pool *p, server_rec *s)
+{
+	scgi_server_cfg *c =
+		(scgi_server_cfg *) ap_pcalloc(p, sizeof(scgi_server_cfg));
+
+	c->mounts = ap_make_array(p, 20, sizeof(mount_entry));
+	return c;
+}
+
+static void *
+merge_server_config(pool *p, void *basev, void *overridesv)
+{
+	scgi_server_cfg *c = (scgi_server_cfg *)
+		ap_pcalloc(p, sizeof(scgi_server_cfg));
+	scgi_server_cfg *base = (scgi_server_cfg *) basev;
+	scgi_server_cfg *overrides = (scgi_server_cfg *) overridesv;
+
+	c->mounts = ap_append_arrays(p, overrides->mounts, base->mounts);
+	return c;
+}
+
+
 static const char *
 cmd_server(cmd_parms *cmd, scgi_cfg *dcfg, char *addr, char *port)
 {
@@ -401,13 +524,13 @@ cmd_server(cmd_parms *cmd, scgi_cfg *dcfg, char *addr, char *port)
 		return "not a server command";
 	} 
 
-	if ((dcfg->addr = inet_addr(addr)) == INADDR_NONE)
+	if ((dcfg->mount.addr = inet_addr(addr)) == INADDR_NONE)
 		return "Invalid syntax for server address";
 
 	n = strtol(port, &tmp, 0);
 	if (tmp[0] != 0 || n < 0 || n > 65535)
 		return "Invalid server port";
-	dcfg->port = htons((unsigned short) n);
+	dcfg->mount.port = htons((unsigned short) n);
 
 	return NULL;
 }
@@ -430,12 +553,41 @@ cmd_handler(cmd_parms *cmd, scgi_cfg *dcfg, int flag)
 	return NULL;
 }
 
+static const char *
+cmd_mount(cmd_parms *cmd, void *dummy, char *path, char *addr)
+{
+	int n;
+	char *colon;
+	char *addr2;
+	char *tmp;
+
+	scgi_server_cfg *scfg = our_sconfig(cmd->server);
+	mount_entry *new = ap_push_array(scfg->mounts);
+	n = strlen(path);
+	while (n > 0 && path[n-1] == '/') {
+		n--; /* strip trailing slashes */
+	}
+	new->path = ap_pstrndup(cmd->pool, path, n);
+	if ((colon = strchr(addr, ':')) == NULL)
+		return "Invalid syntax for server address";
+	addr2 = ap_pstrndup(cmd->pool, addr, colon - addr);
+	if ((new->addr = inet_addr(addr2)) == INADDR_NONE)
+		return "Invalid syntax for server address";
+	n = strtol(colon + 1, &tmp, 0);
+	if (tmp[0] != 0 || n < 0 || n > 65535)
+		return "Invalid server port";
+	new->port = htons((unsigned short) n);
+	return NULL;
+}
+
 static const command_rec scgi_cmds[] =
 {
     { "SCGIServer", cmd_server, NULL, ACCESS_CONF, TAKE2,
       "address and port of an SCGI server (e.g. 127.0.0.1 4000)"},
     { "SCGIHandler", cmd_handler, NULL, ACCESS_CONF, FLAG,
       "On or Off to enable or disable the SCGI handler"},
+    { "SCGIMount", cmd_mount, NULL, RSRC_CONF, TAKE2,
+      "path prefix and address of SCGI server"},
     {NULL}
 };
 
@@ -449,10 +601,10 @@ module scgi_module =
 {
 	STANDARD_MODULE_STUFF,
 	scgi_init,			/* module initializer */
-	scgi_create_dir_config,		/* per-directory config creator */
-	scgi_merge_dir_config,		/* dir config merger */
-	NULL,				/* server config creator */
-	NULL,				/* server config merger */
+	create_dir_config,		/* per-directory config creator */
+	merge_dir_config,		/* dir config merger */
+	create_server_config,		/* server config creator */
+	merge_server_config,		/* server config merger */
 	scgi_cmds,			/* command table */
 	scgi_handlers,			/* [7] list of handlers */
 	scgi_trans,			/* [2] filename-to-URI translation */
