@@ -5,6 +5,7 @@
  */
 
 #include "ap_config.h"
+#include "apr_lib.h"
 #include "apr_strings.h"
 #include "httpd.h"
 #include "http_config.h"
@@ -15,7 +16,7 @@
 #include "mpm_common.h"
 #include "util_script.h"
 
-#define MOD_SCGI_VERSION "1.2"
+#define MOD_SCGI_VERSION "1.7"
 #define SCGI_PROTOCOL_VERSION "1"
 #define DEFAULT_TIMEOUT  60 /* default socket timeout */
 
@@ -23,16 +24,27 @@
 #define ENABLED 1
 #define DISABLED 2
 
+typedef struct {
+    char *path;
+    char *addr;
+    apr_port_t port;
+} mount_entry;
+
 /*
  * Configuration record.  Used per-directory configuration data.
  */
 typedef struct {
-    char *addr;
-    apr_port_t port;
+    mount_entry mount;
     int enabled; /* mod_scgi is enabled from this directory */
     int timeout;
 } scgi_cfg;
 
+/* Server level configuration */
+typedef struct {
+    apr_array_header_t *mounts;
+    int timeout;
+} scgi_server_cfg;
+
 /*
  * Declare ourselves so the configuration routines can find and know us.
  * We'll fill it in at the end of the module.
@@ -48,15 +60,68 @@ our_dconfig(request_rec *r)
     return (scgi_cfg *) ap_get_module_config(r->per_dir_config, &scgi_module);
 }
 
+static scgi_server_cfg *our_sconfig(server_rec *s)
+{
+    return (scgi_server_cfg *) ap_get_module_config(s->module_config,
+                                                    &scgi_module);
+}
+
+static int
+mount_entry_matches(const char *url, const char *prefix,
+                    const char **path_info)
+{
+    int i;
+    for (i=0; prefix[i] != '\0'; i++) {
+        if (url[i] == '\0' || url[i] != prefix[i])
+            return 0;
+    }
+    if (url[i] == '\0' || url[i] == '/') {
+        *path_info = url + i;
+        return 1;
+    }
+    return 0;
+}
+
 static int scgi_translate(request_rec *r)
 {
-	scgi_cfg *cfg = our_dconfig(r);
+    scgi_cfg *cfg = our_dconfig(r);
 
-	if (cfg->enabled == ENABLED) {
-		r->handler = "scgi-handler";
-		return OK;
-	}
-	return DECLINED;
+    if (cfg->enabled == DISABLED) {
+        return DECLINED;
+    }
+
+    if (cfg->mount.addr != UNSET) {
+        ap_assert(cfg->mount.port != UNSET);
+        r->handler = "scgi-handler";
+        r->filename = r->uri;
+        ap_set_module_config(r->request_config, &scgi_module, &cfg->mount);
+        return OK;
+    }
+    else {
+        int i;
+        scgi_server_cfg *scfg = our_sconfig(r->server);
+        mount_entry *entries = (mount_entry *) scfg->mounts->elts;
+        for (i = 0; i < scfg->mounts->nelts; ++i) {
+            const char *path_info;
+            mount_entry *mount = &entries[i];
+            if (mount_entry_matches(r->uri, mount->path, &path_info)) {
+                r->handler = "scgi-handler";
+                r->path_info = apr_pstrdup(r->pool, path_info);
+                r->filename = r->uri;
+                ap_set_module_config(r->request_config, &scgi_module, mount);
+                return OK;
+            }
+        }
+    }
+    return DECLINED;
+}
+
+static int scgi_map_location(request_rec *r)
+{
+    if (r->handler && strcmp(r->handler, "scgi-handler") == 0) {
+        return OK; /* We don't want directory walk. */
+    }
+    return DECLINED;
 }
 
 static void log_err(const char *file, int line, request_rec *r,
@@ -73,6 +138,22 @@ static void log_debug(const char *file, int line, request_rec *r, const
     ap_log_rerror(file, line, APLOG_DEBUG, APR_SUCCESS, r, msg);
 }
 
+static char *http2env(apr_pool_t *p, const char *name)
+{
+    char *env_name = apr_pstrcat(p, "HTTP_", name, NULL);
+    char *cp;
+    
+    for (cp = env_name + 5; *cp != 0; cp++) {
+        if (*cp == '-') {
+            *cp = '_';
+        }
+        else {
+            *cp = apr_toupper(*cp);
+        }
+    }
+
+    return env_name;
+}
 
 static char *lookup_name(apr_table_t *t, const char *name)
 {
@@ -109,6 +190,41 @@ static void add_header(apr_table_t *t, const char *name, const char *value)
         apr_table_addn(t, name, value);
 }
 
+static int find_path_info(const char *uri, const char *path_info)
+{
+    int n;
+    n = strlen(uri) - strlen(path_info);
+    ap_assert(n >= 0);
+    return n;
+}
+
+/* This code is a duplicate of what's in util_script.c.  We can't use
+ * r->unparsed_uri because it gets changed if there was a redirect. */
+static char *original_uri(request_rec *r)
+{
+    char *first, *last;
+
+    if (r->the_request == NULL) {
+	return (char *) apr_pcalloc(r->pool, 1);
+    }
+
+    first = r->the_request;	/* use the request-line */
+
+    while (*first && !apr_isspace(*first)) {
+	++first;		/* skip over the method */
+    }
+    while (apr_isspace(*first)) {
+	++first;		/*   and the space(s)   */
+    }
+
+    last = first;
+    while (*last && !apr_isspace(*last)) {
+	++last;			/* end at next whitespace */
+    }
+
+    return apr_pstrmemdup(r->pool, first, last - first);
+}
+
 /* buffered socket implementation (buckets are overkill) */
 
 #define BUFFER_SIZE 8000
@@ -182,11 +298,12 @@ static apr_status_t bputc(struct sockbuff *s, char c)
 
 
 static apr_status_t
-send_headers(request_rec *r, struct sockbuff *s, scgi_cfg *cfg)
+send_headers(request_rec *r, struct sockbuff *s)
 {
-    apr_table_t *t = apr_table_make(r->pool, 40); /* headers to send */
-    const apr_array_header_t *hdrs_arr;
-    apr_table_entry_t *hdrs;
+    /* headers to send */
+    apr_table_t *t;
+    const apr_array_header_t *hdrs_arr, *env_arr;
+    apr_table_entry_t *hdrs, *env;
     unsigned long int n = 0;
     char *buf;
     int i;
@@ -195,6 +312,9 @@ send_headers(request_rec *r, struct sockbuff *s, scgi_cfg *cfg)
     apr_sockaddr_port_get(&port, r->connection->remote_addr);
 
     log_debug(APLOG_MARK,r, "sending headers");
+    t = apr_table_make(r->pool, 40);
+    if (!t)
+	    return APR_ENOMEM;
     /* CONTENT_LENGTH must come first and always be present */
     add_header(t, "CONTENT_LENGTH",
                apr_psprintf(r->pool, "%ld", r->remaining));
@@ -210,34 +330,38 @@ send_headers(request_rec *r, struct sockbuff *s, scgi_cfg *cfg)
     add_header(t, "REMOTE_PORT", apr_psprintf(r->pool, "%d", port));
     add_header(t, "REMOTE_USER", r->user);
     add_header(t, "REQUEST_METHOD", r->method);
-    add_header(t, "REQUEST_URI", r->unparsed_uri);
+    add_header(t, "REQUEST_URI", original_uri(r));
     add_header(t, "QUERY_STRING", r->args ? r->args : "");
     add_header(t, "SCRIPT_NAME", r->uri);
-    /* skip PATH_INFO, the 'uri' contains the full path */
-    /* add_header(t, "PATH_INFO", r->path_info); */
+    if (r->path_info) {
+        int path_info_start = find_path_info(r->uri, r->path_info);
+        add_header(t, "SCRIPT_NAME", apr_pstrndup(r->pool, r->uri,
+                                                  path_info_start));
+        add_header(t, "PATH_INFO", r->path_info);
+    }
+    else {
+        /* skip PATH_INFO, don't know it */
+        add_header(t, "SCRIPT_NAME", r->uri);
+    }
     add_header(t, "HTTPS", lookup_env(r, "HTTPS"));
     add_header(t, "CONTENT_TYPE", lookup_header(r, "Content-type"));
     add_header(t, "DOCUMENT_ROOT", ap_document_root(r));
-    add_header(t, "HTTP_ACCEPT", lookup_header(r, "Accept"));
-    add_header(t, "HTTP_ACCEPT_CHARSET", lookup_header(r, "Accept-Charset"));
-    add_header(t, "HTTP_ACCEPT_ENCODING", lookup_header(r, "Accept-Encoding"));
-    add_header(t, "HTTP_ACCEPT_LANGUAGE", lookup_header(r, "Accept-Language"));
-    add_header(t, "HTTP_AUTHORIZATION", lookup_header(r, "Authorization"));
-    add_header(t, "HTTP_COOKIE", lookup_header(r, "Cookie"));
-    add_header(t, "HTTP_EXPECT", lookup_header(r, "Expect"));
-    add_header(t, "HTTP_FROM", lookup_header(r, "From"));
-    add_header(t, "HTTP_HOST", lookup_header(r, "Host"));
-    add_header(t, "HTTP_IF_MATCH", lookup_header(r, "If-Match"));
-    add_header(t, "HTTP_IF_MODIFIED_SINCE",
-               lookup_header(r, "If-Modified-Since"));
-    add_header(t, "HTTP_IF_NONE_MATCH", lookup_header(r, "If-None-Match"));
-    add_header(t, "HTTP_IF_RANGE", lookup_header(r, "If-Range"));
-    add_header(t, "HTTP_IF_UNMODIFIED_SINCE",
-               lookup_header(r, "If-Unmodified-Since"));
-    add_header(t, "HTTP_RANGE", lookup_header(r, "Range"));
-    add_header(t, "HTTP_REFERER", lookup_header(r, "Referer"));
-    add_header(t, "HTTP_TE", lookup_header(r, "TE"));
-    add_header(t, "HTTP_USER_AGENT", lookup_header(r, "User-Agent"));
+
+    /* HTTP headers */
+    hdrs_arr = apr_table_elts(r->headers_in);
+    hdrs = (apr_table_entry_t *) hdrs_arr->elts;
+    for (i = 0; i < hdrs_arr->nelts; ++i) {
+        if (hdrs[i].key) {
+            add_header(t, http2env(r->pool, hdrs[i].key), hdrs[i].val);
+        }
+    }
+
+    /* environment variables */
+    env_arr = apr_table_elts(r->subprocess_env);
+    env = (apr_table_entry_t*) env_arr->elts;
+    for (i = 0; i < env_arr->nelts; ++i) {
+        add_header(t, env[i].key, env[i].val);
+    }
 
     hdrs_arr = apr_table_elts(t);
     hdrs = (apr_table_entry_t*) hdrs_arr->elts;
@@ -289,45 +413,56 @@ static apr_status_t send_request_body(request_rec *r, struct sockbuff *s)
     return APR_SUCCESS;
 }
 
+#define CONFIG_VALUE(value, fallback) ((value) != UNSET ? (value) : (fallback))
+
 static apr_status_t
-open_socket(apr_socket_t **sock, request_rec *r, scgi_cfg *cfg)
+open_socket(apr_socket_t **sock, request_rec *r)
 {
-    char *laddr         = cfg->addr ? cfg->addr : "localhost";
-    apr_port_t lport    = cfg->port ? cfg->port : 4000;
-    int ltimeout        = cfg->timeout ? cfg->timeout : DEFAULT_TIMEOUT;
-    apr_interval_time_t timeout = apr_time_from_sec(ltimeout);
+    int timeout;
     int retries = 4;
     int sleeptime = 1;
     apr_status_t rv;
-    apr_sockaddr_t *addr;
-
-    rv = apr_sockaddr_info_get(&addr, laddr, AF_INET, lport, 0, r->pool);
+    apr_sockaddr_t *sockaddr;
+    scgi_server_cfg *scfg = our_sconfig(r->server);
+    scgi_cfg *cfg = our_dconfig(r);
+    mount_entry *m = (mount_entry *) ap_get_module_config(r->request_config,
+                                                          &scgi_module);
+    timeout = CONFIG_VALUE(cfg->timeout, CONFIG_VALUE(scfg->timeout,
+                                                      DEFAULT_TIMEOUT));
+    rv = apr_sockaddr_info_get(&sockaddr,
+                               CONFIG_VALUE(m->addr, "localhost"),
+                               APR_UNSPEC,
+                               CONFIG_VALUE(m->port, 4000),
+                               0,
+                               r->pool);
     if (rv) {
         log_err(APLOG_MARK, r, rv, "apr_sockaddr_info_get() error");
         return rv;
     }
 
+ restart:
     *sock = NULL;
-    rv = apr_socket_create(sock, AF_INET, SOCK_STREAM, r->pool);
+    rv = apr_socket_create(sock, sockaddr->family, SOCK_STREAM, r->pool);
     if (rv) {
         log_err(APLOG_MARK, r, rv, "apr_socket_create() error");
         return rv;
     }
 
-    rv = apr_socket_timeout_set(*sock, timeout);
+    rv = apr_socket_timeout_set(*sock, apr_time_from_sec(timeout));
     if (rv) {
-        log_err(APLOG_MARK, r, rv, "apr_socket_timout_set() error");
+        log_err(APLOG_MARK, r, rv, "apr_socket_timeout_set() error");
         return rv;
     }
 
- restart:
-    rv = apr_socket_connect(*sock, addr);
+    rv = apr_socket_connect(*sock, sockaddr);
     if (rv) {
-        if (APR_STATUS_IS_ECONNREFUSED(rv) && retries > 0) {
+	apr_socket_close(*sock);
+        if ((APR_STATUS_IS_ECONNREFUSED(rv) |
+	     APR_STATUS_IS_EINPROGRESS(rv)) && retries > 0) {
             /* server may be temporarily down, retry */
             ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, rv, r,
-                          "scgi: connection refused, retrying");
-            apr_sleep(sleeptime);
+                          "scgi: connection failed, retrying");
+            apr_sleep(apr_time_from_sec(sleeptime));
             --retries;
             sleeptime *= 2;
             goto restart;
@@ -336,46 +471,41 @@ open_socket(apr_socket_t **sock, request_rec *r, scgi_cfg *cfg)
         return rv;
     }
 
-    ap_sock_disable_nagle(*sock);
+#ifdef APR_TCP_NODELAY
+    /* disable Nagle, we don't send small packets */
+    apr_socket_opt_set(*sock, APR_TCP_NODELAY, 1);
+#endif
+
     return APR_SUCCESS;
 }
 
-
-
 static int scgi_handler(request_rec *r)
 {
     apr_status_t rv = 0;
     int http_status = 0;
-    scgi_cfg *cfg   = NULL;
     struct sockbuff s;
     apr_socket_t *sock;
     apr_bucket_brigade *bb = NULL;
     apr_bucket *b          = NULL;
-    char buf[MAX_STRING_LEN];
     const char *location;
 
     if (strcmp(r->handler, "scgi-handler"))
         return DECLINED;
 
-    cfg = our_dconfig(r);
-    if (!cfg->enabled)
-        return DECLINED;
-
     http_status = ap_setup_client_block(r, REQUEST_CHUNKED_ERROR);
     if (http_status != OK)
         return http_status;
 
     log_debug(APLOG_MARK, r, "connecting to server");
 
-    rv = open_socket(&sock, r, cfg);
+    rv = open_socket(&sock, r);
     if (rv) {
-        log_err(APLOG_MARK, r, rv, "error connecting to scgi server");
         return HTTP_INTERNAL_SERVER_ERROR;
     }
 
     binit(&s, sock);
 
-    rv = send_headers(r, &s, cfg);
+    rv = send_headers(r, &s);
     if (rv) {
         log_err(APLOG_MARK, r, rv, "error sending request headers");
         return HTTP_INTERNAL_SERVER_ERROR;
@@ -400,10 +530,20 @@ static int scgi_handler(request_rec *r)
     b = apr_bucket_eos_create(r->connection->bucket_alloc);
     APR_BRIGADE_INSERT_TAIL(bb, b);
 
-    rv = ap_scan_script_header_err_brigade(r, bb, buf);
+    rv = ap_scan_script_header_err_brigade(r, bb, NULL);
     if (rv) {
-        log_err(APLOG_MARK, r, rv, "error reading response headers");
-        return HTTP_INTERNAL_SERVER_ERROR;
+        if (rv == HTTP_INTERNAL_SERVER_ERROR) {
+            log_err(APLOG_MARK, r, rv, "error reading response headers");
+        }
+        else {
+            /* Work around an Apache bug whereby the returned status is
+             * ignored and status_line is used instead.  This bug is
+             * present at least in 2.0.54.
+             */
+            r->status_line = NULL;
+        }
+        apr_brigade_destroy(bb);
+        return rv;
     }
 
     location = apr_table_get(r->headers_out, "Location");
@@ -444,13 +584,14 @@ static int scgi_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp,
 }
 
 
-static void *scgi_create_dir_config(apr_pool_t *p, char *dirspec)
+static void *
+create_dir_config(apr_pool_t *p, char *dirspec)
 {
     scgi_cfg *cfg = apr_pcalloc(p, sizeof(scgi_cfg));
 
     cfg->enabled = UNSET;
-    cfg->addr = UNSET;
-    cfg->port = UNSET;
+    cfg->mount.addr = UNSET;
+    cfg->mount.port = UNSET;
     cfg->timeout = UNSET;
 
     return cfg;
@@ -459,20 +600,63 @@ static void *scgi_create_dir_config(apr_pool_t *p, char *dirspec)
 #define MERGE(b, n, a) (n->a == UNSET ? b->a : n->a)
 
 static void *
-scgi_merge_dir_config(apr_pool_t *p, void *basev, void *newv)
+merge_dir_config(apr_pool_t *p, void *basev, void *newv)
 {
     scgi_cfg* cfg = apr_pcalloc(p, sizeof(scgi_cfg));
     scgi_cfg* base = basev;
     scgi_cfg* new = newv;
 
     cfg->enabled = MERGE(base, new, enabled);
-    cfg->addr = MERGE(base, new, addr);
-    cfg->port = MERGE(base, new, port);
+    cfg->mount.addr = MERGE(base, new, mount.addr);
+    cfg->mount.port = MERGE(base, new, mount.port);
     cfg->timeout = MERGE(base, new, timeout);
 
     return cfg;
 }
 
+static void *
+create_server_config(apr_pool_t *p, server_rec *s)
+{
+	scgi_server_cfg *c =
+		(scgi_server_cfg *) apr_pcalloc(p, sizeof(scgi_server_cfg));
+
+	c->mounts = apr_array_make(p, 20, sizeof(mount_entry));
+        c->timeout = UNSET;
+	return c;
+}
+
+static void *
+merge_server_config(apr_pool_t *p, void *basev, void *overridesv)
+{
+	scgi_server_cfg *c = (scgi_server_cfg *)
+		apr_pcalloc(p, sizeof(scgi_server_cfg));
+	scgi_server_cfg *base = (scgi_server_cfg *) basev;
+	scgi_server_cfg *overrides = (scgi_server_cfg *) overridesv;
+
+	c->mounts = apr_array_append(p, overrides->mounts, base->mounts);
+        c->timeout = MERGE(base, overrides, timeout);
+	return c;
+}
+
+static const char *
+cmd_mount(cmd_parms *cmd, void *dummy, const char *path, const char *addr)
+{
+    int n;
+    apr_status_t rv;
+    char *scope_id = NULL; /* A ip6 parameter - not used here. */
+    scgi_server_cfg *scfg = our_sconfig(cmd->server);
+    mount_entry *new = apr_array_push(scfg->mounts);
+    n = strlen(path);
+    while (n > 0 && path[n-1] == '/') {
+        n--; /* strip trailing slashes */
+    }
+    new->path = apr_pstrndup(cmd->pool, path, n);
+    rv = apr_parse_addr_port(&new->addr, &scope_id, &new->port, addr,
+                             cmd->pool);
+    if (rv)
+        return "error parsing address:port string";
+    return NULL;
+}
 
 static const char *
 cmd_server(cmd_parms *cmd, void *pcfg, const char *addr_and_port)
@@ -484,7 +668,7 @@ cmd_server(cmd_parms *cmd, void *pcfg, const char *addr_and_port)
     if (cmd->path == NULL)
         return "not a server command";
 
-    rv = apr_parse_addr_port(&cfg->addr, &scope_id, &cfg->port,
+    rv = apr_parse_addr_port(&cfg->mount.addr, &scope_id, &cfg->mount.port,
                              addr_and_port, cmd->pool);
     if (rv)
         return "error parsing address:port string";
@@ -513,24 +697,29 @@ cmd_handler(cmd_parms* cmd, void* pcfg, int flag)
 static const char *
 cmd_timeout(cmd_parms *cmd, void* pcfg, const char *strtimeout)
 {
-    scgi_cfg *cfg = pcfg;
+    scgi_cfg *dcfg = pcfg;
+    int timeout = atoi(strtimeout);
 
-    if (cmd->path == NULL)
-        return "not a server command";
-
-    cfg->timeout = atoi(strtimeout);
+    if (cmd->path == NULL) {
+        scgi_server_cfg *scfg = our_sconfig(cmd->server);
+        scfg->timeout = timeout;
+    }
+    else {
+        dcfg->timeout = timeout;
+    }
 
     return NULL;
 }
 
-
 static const command_rec scgi_cmds[] =
 {
+    AP_INIT_TAKE2("SCGIMount", cmd_mount, NULL, RSRC_CONF,
+                  "path prefix and address of SCGI server"),
     AP_INIT_TAKE1("SCGIServer", cmd_server, NULL, ACCESS_CONF,
                   "Address and port of an SCGI server (e.g. localhost:4000)"),
     AP_INIT_FLAG( "SCGIHandler", cmd_handler, NULL, ACCESS_CONF,
                   "On or Off to enable or disable the SCGI handler"),
-    AP_INIT_TAKE1("SCGIServerTimeout", cmd_timeout, NULL, ACCESS_CONF,
+    AP_INIT_TAKE1("SCGIServerTimeout", cmd_timeout, NULL, ACCESS_CONF|RSRC_CONF,
                   "Timeout (in seconds) for communication with the SCGI server."),
     {NULL}
 };
@@ -541,16 +730,17 @@ static void scgi_register_hooks(apr_pool_t *p)
     ap_hook_post_config(scgi_init, NULL, NULL, APR_HOOK_MIDDLE);
     ap_hook_handler(scgi_handler, NULL, NULL, APR_HOOK_MIDDLE);
     ap_hook_translate_name(scgi_translate, NULL, NULL, APR_HOOK_LAST);
+    ap_hook_map_to_storage(scgi_map_location, NULL, NULL, APR_HOOK_FIRST);
 }
 
 
 /* Dispatch list for API hooks */
 module AP_MODULE_DECLARE_DATA scgi_module = {
     STANDARD20_MODULE_STUFF,
-    scgi_create_dir_config,             /* create per-dir config structs */
-    scgi_merge_dir_config,              /* merge per-dir config structs */
-    NULL,                               /* create per-server config structs */
-    NULL,                               /* merge per-server config structs */
-    scgi_cmds,                          /* table of config file commands */
-    scgi_register_hooks,                /* register hooks */
+    create_dir_config,             /* create per-dir config structs */
+    merge_dir_config,              /* merge per-dir config structs */
+    create_server_config,          /* create per-server config structs */
+    merge_server_config,           /* merge per-server config structs */
+    scgi_cmds,                     /* table of config file commands */
+    scgi_register_hooks,           /* register hooks */
 };
